@@ -11,19 +11,22 @@ from itertools import chain, islice, dropwhile, takewhile, starmap
 from functools import reduce
 from typing import Callable, Union, List, Set, Iterator, Iterable, TypeVar, Generic, Dict, Tuple, Any
 from .util import to_iterator
-from .operator import Deduplicator, Inserter
+from .operator import Deduplicator, Inserter, PairUp, Zipper, Collapser, Grouper
 from .collector import Collector, CountCollector
 
 T = TypeVar('T')
 R = TypeVar('R')
 ElementOrIter = Union[Iterable[T], Iterator[T], T]
 
+K = TypeVar('K')
+V = TypeVar('V')
+
 
 class Stream(Generic[T]):
     """
     The Stream class is the chainable wrapper class around any generators / iterators
     Generators and iterators are merged together, and the sequence is preserved - the same at creation time.
-    No evaluation will actually happen unless a consumer operation is executed.
+    No evaluation will actually happen unless a terminal / near-terminal operation is executed.
     """
 
     @staticmethod
@@ -147,6 +150,9 @@ class Stream(Generic[T]):
         :param exclusion: items to exclude
         :return: New stream with exclusion filtered
         """
+        if len(exclusion) == 0:
+            return self
+
         all_exclusions = set(exclusion)
         return Stream(elem for elem in self.__stream if elem not in all_exclusions)
 
@@ -167,12 +173,12 @@ class Stream(Generic[T]):
         return Stream((maybe_elem,), self)
 
     ###
-    # Consumer operations
+    # Terminal operations
     ###
 
     def collect(self, collector: Union[Callable[[Iterator[T]], R], Collector[T, Any, R]]) -> R:
         """
-        [Consumer operation] passes the stream to collector
+        [Terminal operation] passes the stream to collector
         :param collector: (Stream -> any) function iterates through the stream
         :return: the result after piping stream to collector function
         """
@@ -185,29 +191,29 @@ class Stream(Generic[T]):
 
     def collect_as_list(self) -> List[T]:
         """
-        [Consumer operation] convert to a list
+        [Terminal operation] convert to a list
         :return: List
         """
         return list(self)
 
     def collect_as_set(self) -> Set[T]:
         """
-        [Consumer operation] convert to a set
+        [Terminal operation] convert to a set
         :return: Set
         """
         return set(self)
 
     def collect_dict(self, dict_collector: Callable[[Iterator[T]], Dict] = dict):
         """
-        [Consumer operation] form a map/dict with the 1st element from each stream candidate as keys and rest as values
+        [Terminal operation] form a map/dict with the 1st element from each stream candidate as keys and rest as value
         :param dict_collector: (Stream<I extends map.entry> -> Dict<K, V>) function iterates through the stream
         :return: the result after piping stream to dict collector function
         """
-        return dict_collector((elem[0], elem[1] if len(elem) == 2 else elem[1:]) for elem in self.__stream)
+        return dict_collector(self.wrap_as_dict_stream())
 
     def build_dict(self, dict_collector: Callable[[Iterator[T]], Dict] = dict):
         """
-        (alias of collect_dict) [Consumer operation] passes the stream to dict collector
+        (alias of collect_dict) [Terminal operation] passes the stream to dict collector
         :param dict_collector: (stream<I extends item> -> D extends dict) function iterates through the item stream
         :return: the result after piping stream to dict collector function
         """
@@ -215,7 +221,7 @@ class Stream(Generic[T]):
 
     def collect_as_map(self, map_collector: Callable[[Iterator[T]], Dict] = dict):
         """
-        (alias of collect_dict) [Consumer operation] passes the stream to dict collector
+        (alias of collect_dict) [Terminal operation] passes the stream to dict collector
         :param map_collector: (stream<I extends item> -> D extends dict) function iterates through the item stream
         :return: the result after piping stream to dict collector function
         """
@@ -223,10 +229,11 @@ class Stream(Generic[T]):
 
     def reduce(self, reducer: Callable[[R, T], R], initial_value: Union[R, None] = None):
         """
-        [Consumer operation] equivalent to functool reduce. Passes the stream to reducer
+        [Terminal operation] equivalent to functool reduce. Passes the stream to reducer
         :param reducer: (T extends any, element -> T) function takes each item and produce an (aggregated) value
         :param initial_value: (T) optional value, served as the starting value.
         :return: the result after reducing the stream
+        TODO: short circuitable reduction
         """
         if initial_value is not None:
             return reduce(reducer, self.__stream, initial_value)
@@ -235,7 +242,7 @@ class Stream(Generic[T]):
 
     def reduce_right(self, reducer: Callable[[R, T], R], initial_value: Union[R, None] = None) -> R:
         """
-        [Consumer operation] equivalent to functool reduce. Passes the stream in reverse order to reducer
+        [Terminal operation] equivalent to functool reduce. Passes the stream in reverse order to reducer
         :param reducer: (R result, T element -> T) function takes each item and produce an (aggregated) value
         :param initial_value: (R) optional value, served as the starting value.
         :return: R - the result after reducing the stream
@@ -248,7 +255,7 @@ class Stream(Generic[T]):
 
     def foreach(self, func: Callable[[T], None]) -> None:
         """
-        [Consumer operation] passes the stream to func
+        [Terminal operation] passes the stream to func
         :param func: (element -> void) function runs on each element
         """
         for element in self:
@@ -256,7 +263,7 @@ class Stream(Generic[T]):
 
     def foreach_index(self, func: Callable[[int, T], None]) -> None:
         """
-        [Consumer operation] passes the stream to func
+        [Terminal operation] passes the stream to func
         :param func: (element -> void) function runs on each element
         """
         for i, element in enumerate(self):
@@ -264,7 +271,7 @@ class Stream(Generic[T]):
 
     def any_match(self, match: Callable[[T], bool]) -> bool:
         """
-        [Consumer operation] test if any element in the stream passes the predicate test
+        [Terminal operation] test if any element in the stream passes the predicate test
         :param match: (element -> bool) function tests each element
         :return: test result
         """
@@ -273,9 +280,32 @@ class Stream(Generic[T]):
                 return True
         return False
 
+    def has_all(self, *candidate: T) -> bool:
+        """
+        [Terminal operation] test if all candidates exist in the stream
+        :param candidate: candidates to check
+        :return: test result
+        """
+        candidates = set(candidate)
+        for item in self.__stream:
+            if item in candidates:
+                candidates.remove(item)
+            if len(candidates) == 0:
+                return True
+        return False
+
+    def has_any(self, *candidate: T) -> bool:
+        """
+        [Terminal operation] test if at least one of the candidates exist in the stream
+        :param candidate: candidates to check
+        :return: test result
+        """
+        candidates = set(candidate)
+        return self.any_match(lambda item: item in candidates)
+
     def all_match(self, match: Callable[[T], bool]) -> bool:
         """
-        [Consumer operation] test if all elements in the stream pass the predicate test
+        [Terminal operation] test if all elements in the stream pass the predicate test
         :param match: (element -> bool) function tests each element
         :return: test result
         """
@@ -286,7 +316,7 @@ class Stream(Generic[T]):
 
     def none_match(self, match: Callable[[T], bool]) -> bool:
         """
-        [Consumer operation] test if none of elements in the stream pass the predicate test
+        [Terminal operation] test if none of elements in the stream pass the predicate test
         :param match: (element -> bool) function tests each element
         :return: test result
         """
@@ -294,29 +324,37 @@ class Stream(Generic[T]):
 
     def count(self) -> int:
         """
-        [Consumer operation] count total number of elements
+        [Terminal operation] count total number of elements
         :return: total number
         """
         return self.collect(CountCollector())
 
     def sorted(self, key: Union[Callable[[T], Any], None] = None, reverse: bool = False):
         """
-        [Consumer operation] effectively collects all element for comparison and sorting for a sorted stream
-        :param: key - optional comparison method
-        :param: reverse - if the order of sort should be reversed (decreasing order)
+        [Near terminal operation] effectively collects all element for comparison and sorting for a sorted stream
+        :param key: optional comparison method
+        :param reverse: if the order of sort should be reversed (decreasing order)
         :return: A sorted stream
         TODO: stream sort condition marker
         TODO: parallel implementation
         """
         return Stream(sorted(self, key=key, reverse=reverse))
 
+    def for_pairs(self, func: Callable[[Tuple[T, T]], None]) -> None:
+        """
+        [Terminal operation] apply function to every adjacent pair
+        :param func: function applying on all adjacent pairs
+        """
+        for pair in PairUp(self.__stream):
+            func(pair)
+
     ###
     # Element extraction method
     ###
 
     def find_any(self) -> Union[T, None]:
         """
-        [Extraction operation] get any one element in the stream
+        [Terminal operation] get any one element in the stream
         :return: Optional[T] - maybe element
         """
         try:
@@ -326,7 +364,7 @@ class Stream(Generic[T]):
 
     def find_first(self) -> Union[T, None]:
         """
-        [Extraction operation] get first element in the stream
+        [Terminal operation] get first element in the stream
         The same as find_any in single thread stream
         :return: Optional[T] - maybe element
         """
@@ -339,8 +377,8 @@ class Stream(Generic[T]):
     def distinct(self, *, more_than: int = 1, key: Union[Callable[[T], Any], None] = None):
         """
         Gives stream with distinct elements
-        :param: more_than - only show elements that appear at least this number of times; at least 1
-        :param: key - apply this function to the element for de-duplicating; \
+        :param more_than - only show elements that appear at least this number of times; at least 1
+        :param key - apply this function to the element for de-duplicating; \
             only first of the elements have same key will be preserved.
         :return: stream with distinct elements
         TODO: stream distinct condition marker
@@ -405,7 +443,7 @@ class Stream(Generic[T]):
 
     def max(self, key: Union[Callable[[T], Any], None] = None):
         """
-        [Consumer operation] grab the max value in the stream
+        [Terminal operation] grab the max value in the stream
         :param key: (element -> C extends comparable) optional evaluator to compare elements
         :return: the max element in the stream
         """
@@ -415,7 +453,7 @@ class Stream(Generic[T]):
 
     def min(self, key: Union[Callable[[T], Any], None] = None):
         """
-        [Consumer operation] grab the min value in the stream
+        [Terminal operation] grab the min value in the stream
         :param key: (element -> C extends comparable) optional evaluator to compare elements
         :return: the min element in the stream
         """
@@ -425,12 +463,131 @@ class Stream(Generic[T]):
 
     def intersperse(self, delimiter: T):
         """
-        A new stream in which the delimiter is inserted in between every adjacent elements.
+        Creates a new stream in which the delimiter is inserted in between every adjacent elements.
         :param delimiter: same kind of existing elements
         :return: a new stream
         """
         return Stream(Inserter(self, delimiter))
 
+    def cross(self, elements: Iterable[R]):
+        """
+        Creates a new stream contain every pair of values and the cross elements.
+        :param elements: for cross product
+        :return: a new stream
+        """
+        frozen_elements = frozenset(elements)
+        return Stream((item, elem) for item in self.__stream for elem in frozen_elements)
+
+    def cross_result_of(self, generate: Callable[[T], Iterator[R]]):
+        """
+        Creates a new stream contain every pair of values and the results of generate function.
+        :param generate: generator function
+        :return: a new stream
+        """
+        return Stream((item, value) for item in self.__stream for value in generate(item))
+
+    def map_pairs(self, func: Callable[[Tuple[T, T]], R]):
+        """
+        Maps to new item on every adjacent pair
+        :param func: mapping function applying on all adjacent pairs
+        """
+        return Stream(func(pair) for pair in PairUp(self.__stream))
+
+    def map_to_entry(self, to_entry: Callable[[T], Tuple[K, V]]):
+        """
+        Creates a dict stream, entries of which are created by applying `to_entry` to existing element.
+        :param to_entry: function to create K, V tuple pair from existing elements
+        :return: DictStream
+        """
+        return DictStream(wrap=map(to_entry, self.__stream))
+
+    def map_to_key_value(self, to_key: Callable[[T], K], to_val: Callable[[T], V]):
+        """
+        Creates a dict stream, entries of which are key-value pairs by applying `to_key` `to_val` to existing element,
+        respectively.
+        :param to_key: function to create key from existing elements
+        :param to_val: function to create val from existing elements
+        :return: DictStream
+        """
+        return DictStream(wrap=((to_key(elem), to_val(elem)) for elem in self.__stream))
+
+    def zip_with(self, *streams: Iterator[R], fill_none: bool = False):
+        """
+        Creates a new stream, the i-th element of which is a tuple of i-th elements of this plus all other streams.
+        :param streams: Other streams to get zipped
+        :param fill_none: True - stop fast when one of the stream depletes;
+            False - fill in None for depleted streams until all streams deplete.
+        :return: Stream with zipped tuples
+        """
+        if len(streams) == 0:
+            return self
+        return Stream(Zipper(self.__stream, *streams, fill_none=fill_none))
+
+    def collapse_to_first(self, collapsible: Callable[[T, T], bool]):
+        """
+        Creates a stream with only the leading elements of all consecutively collapsible element chains
+        :param collapsible: determine if a pair of adjacent elements is collapsible.
+        :return: stream with collapsed result
+        """
+        return Stream(Collapser(self.__stream, collapsible, collector=Collector.of(lambda l: l[0])))
+
+    def collapse_and_combine(self, collapsible: Callable[[T, T], bool], combiner: Callable[[T, T], T]):
+        """
+        Creates a stream with all consecutively collapsible element chains combined with `combiner`
+        :param collapsible: determine if a pair of adjacent elements is collapsible.
+        :param combiner: apply this rule to the final chain of collapsible elements
+        :return: stream with collapsed result
+        """
+        return Stream(Collapser(self.__stream, collapsible, combiner=combiner))
+
+    def collapse_and_collect(self, collapsible: Callable[[T, T], bool], collector: Collector):
+        """
+        Creates a stream with all consecutively collapsible element chains collected `collector`
+        :param collapsible: determine if a pair of adjacent elements is collapsible.
+        :param collector: apply this collector to the final chain of collapsible elements
+        :return: stream with collapsed result
+        """
+        return Stream(Collapser(self.__stream, collapsible, collector=collector))
+
+    def adjacent_groups(self, in_same_group: Callable[[T, T], bool]):
+        """
+        Creates a stream of list with adjacent `in_same_group` elements groupped in
+        :param in_same_group: determine if an adjacent pair still in the same group.
+        :return: stream with adjacent element groups
+        """
+        return Stream(Collapser(self.__stream, in_same_group, collector=Collector.of(list)))
+
+    def adjacent_key_groups(self, group_key_of: Callable[[T], R]):
+        """
+        Creates a stream of tuples of group key / adjacent elements
+        :param group_key_of: get the group key of an element
+        :return: stream with group tuples
+        """
+        return Stream(Collapser(
+            self.__stream,
+            collapsible=lambda x, y: group_key_of(x) == group_key_of(y),
+            collector=Collector.of(lambda l: (group_key_of(l[0]), l))))
+
+    def group_by(self, key: Callable[[T], K]):
+        """
+        Creates a stream of map entries, keys of which are the result by applying `key` on all elements, and values of
+        which are the list of elements that result in the same key.
+        :param key: key generating function
+        :return: stream of map entries
+        """
+        return DictStream(wrap=Grouper(self.map_to_key_value(key, lambda item: item)))
+
+    def group_to_map(
+            self, key: Callable[[T], K], *, map_collector: Callable[[Iterator[Tuple]], Dict] = dict) -> Dict[K, List[T]]:
+        """
+        Creates a map, keys of which are the result by applying `key` on all elements, and values of which are
+        the list of elements that result in the same key.
+        :param key: key generating function
+        :param map_collector: transforming entry stream to a dict
+        :return: dict
+        """
+        return self.group_by(key).collect_as_map(map_collector=map_collector)
+
     def stream_transform(self, stream_func: Callable[[Iterator[T]], Iterator[R]]):
         """
         Run arbitrary stream transformation
@@ -439,9 +596,15 @@ class Stream(Generic[T]):
         """
         return Stream(stream_func(self.__stream))
 
-
-K = TypeVar('K')
-V = TypeVar('V')
+    def wrap_as_dict_stream(self):
+        """
+        Try box up current stream as a dict stream. Use at your own risk if the stream content is not in a format of
+        two-element tuple.
+        :return: DictStream
+        """
+        if isinstance(self, DictStream):
+            return self
+        return DictStream(wrap=((elem[0], elem[1] if len(elem) == 2 else elem[1:]) for elem in self.__stream))
 
 
 class DictStream(Stream[Tuple[K, V]]):
